/*+------------------------------------------------------+
  |        simplePing.c  -  description                  |
  |            -------------------                       |
  | begin      : 06/02/2011 11.39                        |
  | copyright  : (C) 2011 xAppSoftware                   |
  | author     : Luigi D'Andrea                          |
  | email      : gg1 at xappsoftware.com                 |
  | compiling  : make                                    |
  |                                                      |
  | Latest version on http://www.xappsoftware.com        |
  +------------------------------------------------------+*/

/*+------------------------------------------------------+
  | SYSTEM INCLUDES                                      |
  +------------------------------------------------------+*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <resolv.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip_icmp.h>

/*+------------------------------------------------------+
  | SPECIFIC INCLUDES                                    |
  +------------------------------------------------------+*/
#include "simplePing.h"

/*+------------------------------------------------------+
  | GLOBAL VARIABLES                                     |
  +------------------------------------------------------+*/
globalData_t globalData;
int          done             =  0;
int          donel            =  0;
int          packetsSent      =  0;
int          packetsReceived  =  0;
int          appo             =  0;

/*+------------------------------------------------------+
  | Function name : checksum                             |
  | Parameters    :                                      |
  |        in     : buffer and its lenght.               |
  | Return  value : returns the computed checksum.       |
  | Description   : Standard 1s complement checksum      |
  +------------------------------------------------------+*/
unsigned short checksum(void *b, int len)
{
	unsigned short   *buf = b;
	unsigned int     sum=0;

	for(sum=0; len>1; len-=2)
	{
		sum += *buf++;
	}
	if (len == 1)
	{
		sum += *(unsigned char*)buf;
	}
	sum     =  (sum >> 16) + (sum & 0xFFFF);
	sum     += (sum >> 16);
	return(~sum);
}

/*+------------------------------------------------------+
  | Function name : printInfo                            |
  | Parameters    :                                      |
  |        in     :                                      |
  | Return  value :                                      |
  | Description   : Prints data on the screen.           |
  +------------------------------------------------------+*/
void printInfo(void *buf, int bytes)
{
	int                i;
	struct iphdr       *ip    =  buf;
	struct icmphdr     *icmp  =  buf+ip->ihl*4;

	if(globalData.verbose==1)
	{
		for(i=0; i<bytes; i++)
		{
			if (!(i & 15))
				printf("\n0x%02X:  ", i);
				printf("%02X ", ((unsigned char*)buf)[i]);
		}
		printf("\n");
	}
	if(strcmp(inet_ntoa(*(struct in_addr *)&ip->daddr), inet_ntoa(*(struct in_addr *)&ip->saddr))==0)
	{
		// a ping on 127.0.0.1 (localhost) creates 2 responses, one from
		// the kernel and the second one from this program, to avoid this behaviour
		// we remove the first one.
		if(appo&1)
		{
			printf("IPv%d: hdr-size=%d pkt-size=%d protocol=%d TTL=%d src=%s ",
				ip->version, ip->ihl*4, ntohs(ip->tot_len), ip->protocol,
				ip->ttl, inet_ntoa(*(struct in_addr *)&ip->saddr));
			printf("dst=%s\n", inet_ntoa(*(struct in_addr *)&ip->daddr));
			packetsReceived++;
			if (icmp->un.echo.id == globalData.pid)
			{
				printf("ICMP: type[%d/%d] checksum[%d] id[%d] seq[%d]\n",
					icmp->type, icmp->code, ntohs(icmp->checksum),
					icmp->un.echo.id, icmp->un.echo.sequence);
			}
		}
		appo++;
	}
}

/*+------------------------------------------------------+
  | Function name : listener                             |
  | Parameters    :                                      |
  |        in     :                                      |
  | Return  value :                                      |
  | Description   : Listen for returning packets.        |
  +------------------------------------------------------+*/
void *listener(void *par)
{
	int                  sd;
	struct sockaddr_in   addr;
	unsigned char        buf[1024];
	socklen_t            len;
	int                  bytes;

	len  =  sizeof(struct sockaddr_in);
	sd   =  socket(PF_INET, SOCK_RAW, globalData.protocol->p_proto);
	if(sd<0)
	{
		perror("ping:socket");
		exit(0);
	}
	while(!donel)
	{
		bzero(buf, sizeof(buf));
		bytes = recvfrom(sd, buf, sizeof(buf), 0, (struct sockaddr*)&addr, &len);
		if (bytes > 0)
		{
			printInfo(buf, bytes);
		}
		else
		{
			perror("ping:recvfrom");
		}
	}
	return(NULL);
}

/*+------------------------------------------------------+
  | Function name : ping                                 |
  | Parameters    :                                      |
  |        in     : The destination IP address.          |
  | Return  value : NULL or EXIT_FAILURE.                |
  | Description   : The pinger function.                 |
  +------------------------------------------------------+*/
void *ping(void *par)
{
	const int            val  =   255;
	int                  i, sd;
	struct packet        pckt;
	socklen_t            len;
	struct sockaddr_in   *addr;

	addr  =  (struct sockaddr_in *)par;
	len   =  sizeof(struct sockaddr_in);
	sd    =  socket(PF_INET, SOCK_RAW, globalData.protocol->p_proto);
	if (sd < 0)
	{
		perror("socket");
		exit(EXIT_FAILURE);
	}
	if (setsockopt(sd, SOL_IP, IP_TTL, &val, sizeof(val)) != 0)
	{
		perror("Set TTL option");
		exit(EXIT_FAILURE);
	}
	if (fcntl(sd, F_SETFL, O_NONBLOCK) != 0)
	{
		perror("Request nonblocking I/O");
		exit(EXIT_FAILURE);
	}
	while(!done)
	{
		printf("Msg #%d\n", packetsSent);
		bzero(&pckt, sizeof(pckt));
		pckt.hdr.type               =  ICMP_ECHO;
		pckt.hdr.un.echo.id         =  globalData.pid;
		for (i = 0; i < sizeof(pckt.msg)-1; i++)
			pckt.msg[i] = i+'0';
		pckt.msg[i]                 =  0;
		pckt.hdr.un.echo.sequence   =  packetsSent++;
		pckt.hdr.checksum           =  checksum(&pckt, sizeof(pckt));
		if (sendto(sd, &pckt, sizeof(pckt), 0, (struct sockaddr*)addr, sizeof(*addr)) <= 0)
		{
			perror("sendto");
		}
		sleep(1);
	}
	return(NULL);
}

/*+------------------------------------------------------+
  | Function name : printStatistics                      |
  | Parameters    :                                      |
  |        in     : None.                                |
  | Return  value : None.                                |
  | Description   : prints the ping statistics.          |
  +------------------------------------------------------+*/
void printStatistics(void)
{
	printf("Packets sent      : %d\n", packetsSent);
	printf("Packets received  : %d\n", packetsReceived);
}

/*+------------------------------------------------------+
  | Function name : quitProgram                          |
  | Parameters    :                                      |
  |        in     : The received signal.                 |
  | Return  value : None.                                |
  | Description   : Catches the CTRL-C, prints the stats |
  |                 and closes both the threads.         |
  +------------------------------------------------------+*/
void quitProgram(int sig)
{
	printf("\n\nExiting simplePing\n");
	donel = 1;
	sleep(1);
	done  = 1;
	printStatistics();
}

/*+------------------------------------------------------+
  | Function name : main                                 |
  | Parameters    :                                      |
  |        in     : The destination IP address.          |
  | Return  value : EXIT_SUCCES or EXIT_FAILURE.         |
  | Description   : The main function for the ping.      |
  +------------------------------------------------------+*/
int main(int argc, char **argv) /* REQ_OS_03 CLI arguments.   */
{
	pthread_t           listenerThreadId;
	pthread_t           pingerThreadId;
    struct sockaddr_in  *addr;

	/* Trapping the CTRL-C                                    */
	signal(SIGINT, quitProgram);
    if ((argc<2)||(argc>3))     /* REQ_OS_03 CLI arguments.   */
        Usage(argv);

	/* Initialize global data                                 */
	addr = initializeGlobalData(argc, argv);

	/* REQ_OS_04.a                                            */
	/* Create the listener thread                             */
	pthread_create(&listenerThreadId, NULL, listener, NULL);

	sleep(1);
	/* REQ_OS_04.b                                            */
	/* Create the pinger thread                               */
	pthread_create(&pingerThreadId, NULL, ping, addr);

	/* Wait for the end of both threads.                      */
	pthread_join(pingerThreadId, NULL);
	pthread_join(listenerThreadId, NULL);

    exit(EXIT_SUCCESS);
}

/*+------------------------------------------------------+
  | Function name : Usage                                |
  | Parameters    :                                      |
  |        in     : The Command Line Arguments.          |
  | Return  value : EXIT_FAILURE.                        |
  | Description   : Prints intrsuctions for the use      |
  +------------------------------------------------------+*/
void Usage(char **argv)
{
    printf("Usage: %s <addr> [-v]\n", argv[0]);
    printf("\tWhere <addr> is the IP address to ping\n");
    printf("\tThe -v option enables verbose output.\n");
    exit(EXIT_FAILURE);
}

/*+------------------------------------------------------+
  | Function name : initializeGlobalData                 |
  | Parameters    :                                      |
  |        in     : None.                                |
  | Return  value : None.                                |
  | Description   : Initializes the global data structure|
  +------------------------------------------------------+*/
struct sockaddr_in *initializeGlobalData(int argc, char **argv)
{
    struct hostent      *hostname;
    struct sockaddr_in  *addr;

	/* Setup the address                                      */
	addr = (struct sockaddr_in*)malloc(sizeof(struct sockaddr_in));
	if(addr==NULL)
	{
		perror("Cannot allocate memory for address");
		exit(EXIT_FAILURE);
	}
    bzero(addr, sizeof(addr));
    hostname                =  gethostbyname(argv[1]);
    addr->sin_family        =  hostname->h_addrtype;
    addr->sin_port          =  0;
    addr->sin_addr.s_addr   =  *(long*)hostname->h_addr;

	/* Setup the current pid                                  */
    globalData.pid         =  getpid();

	/* Setup the protocol                                     */
    globalData.protocol    =  getprotobyname("ICMP");

	/* Check for the verbose flag                             */
	if(argc==3)
	{
		if(strstr(argv[2], "-v")!=NULL)
		{
			globalData.verbose = 1;
			printf("verbose mode on\n");
		}
		else
		{
			globalData.verbose = 0;
		}
	}
	else
		globalData.verbose = 0;
	
	return(addr);
}

